/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */

#define FM_DEFINE_GLOBALS
#include "fm.h"

static void init_globals();
static int process_args(int current_arg, int arg_count);
static void show_usage_message();
static void set_vp_to_manager_mapping_type();
static void node_main();
static void send_initialization_to_nodes();
static void recv_initialization_from_master();

#ifdef DEBUG
extern int _fm_compiled_with_debugging;
extern int _yoke_compiled_with_debugging;
#else 
extern int _fm_not_compiled_with_debugging;
extern int _yoke_not_compiled_with_debugging;
#endif
#ifdef MALLOC_DEBUG
extern int _fm_compiled_with_malloc_debugging;
extern int _yoke_compiled_with_malloc_debugging;
#else 
extern int _fm_not_compiled_with_malloc_debugging;
extern int _yoke_not_compiled_with_malloc_debugging;
#endif
#ifdef PROFILING
extern int _fm_compiled_with_profiling;
extern int _yoke_compiled_with_profiling;
#else
extern int _fm_not_compiled_with_profiling;
extern int _yoke_not_compiled_with_profiling;
#endif

/*
 * FM_INIT()	Fortran Interface Function
 *
 * The machine size argument is generated from the value of the
 * main program's PROCESSORS declaration. 
 */
void FM_INIT(f_integer *dims, f_integer *n_dims)
{
    int child_pid;
    conn_t *node_list;
    int n_nodes;
    conn_t him_conn, sender;
    yoke_buffer_t buf;
    int type, who, i;

    init_globals();
    
    /*
     * Initialize the transport layer. All nodes in the computation
     * will "return" from yoke_init().
     */
    yoke_init("-fm",
	      process_args,
	      (int (*)(char *, int)) NULL,
	      show_usage_message,
	      &n_nodes, &node_list);

#ifdef DEBUG
    if (ProcessDebug(2))
	yoke_printf("FM Process running self=%s\n",
		    yoke_conn_as_string(yoke_self()));
#endif

    _fm_register_handlers();

    if (n_nodes == 0)
    {
	/*
	 * We are a remote node. Drop into that code and never
	 * return.
	 */
	node_main();
	exit(0);
    }

    /*
     * We're the master. Send off any vital information to the rest
     * of the managers, and start up the process for the user code.
     */

    _fm_machine_size = dims[0];
    Assert(_fm_machine_size > 0 && _fm_machine_size < 100000);
    _fm_manager_count = n_nodes;
    _fm_manager_list = node_list;

    _fm_manager_id = 0;

    set_vp_to_manager_mapping_type();
    
    send_initialization_to_nodes();

    him_conn = yoke_fork(&child_pid);

    if (child_pid == 0)
    {
	/*
	 * The child initializes the per-process state and returns from
	 * fm_init() to execute user code.
	 *
	 * The arguments to fm_process_init() are the (virtual) processor
	 * id for this child and the child's offset into the virtual
	 * computer table. Both start out at zero.
	 */

	_fm_process_name = (char *) malloc(strlen("main") + 11);
	strcpy(_fm_process_name, "main");
	
	_fm_parent_process_name = (char *) malloc(strlen("toplevel") + 1);
	strcpy(_fm_parent_process_name, "toplevel");

	_fm_process_init(0, 0);

	/*
	 * Wait for our manager to tell us it is ok to continue.
	 */
	StartSafeRegion;
	buf = yoke_recv_transient(&sender, &type, FALSE);
	if (type != FM_MTYPE_MAIN_RUN)
	    yoke_fatal("fm_init(): Wanted FM_MTYPE_MAIN_RUN, but got %d instead\n",
		       type);
	yoke_get_conn(buf, &_fm_my_manager);
	yoke_free_buffer(buf);
	yoke_release_transient(sender);
	EndSafeRegion;

#ifdef DEBUG
	if (FMDebug(2))
	    yoke_printf("fm_init() returning to user code\n");

#endif
	   
	if (_fm_pause_on_user_startup)
	{
	    yoke_printf("fm_init(): Pausing on user startup\n");
	    pause();
	}
	else if (_fm_sleep_on_user_startup > 0)
	{
	    yoke_printf("fm_init(): Sleeping %d seconds on user startup\n",
			_fm_sleep_on_user_startup);
	    sleep(_fm_sleep_on_user_startup);
	}

	return;
    }

    /*
     * Otherwise we are the manager. Initialize, set up the child's entry
     * in our tables, tell the child to go ahead and execute,
     * and drop into the manager main code. We never
     * return from fm_init().
     *
     * We pass our conn_t to the child so he has a handle on the world.
     */

    _fm_manager_init();
    _fm_manager_add_child(child_pid, him_conn);

    /*
     * Wait for our nodes to acknowledge their readiness to compute.
     */

    for (i = 0; i < _fm_manager_count - 1; i++)
    {
	StartSafeRegion;
	buf = yoke_recv_transient(&sender, &type, FALSE);
	if (type != FM_MTYPE_MANAGER_READY)
	    yoke_fatal("fm_init(): Wanted FM_MTYPE_MANAGER_READY, but got %d instead\n",
		       type);
	yoke_get_int(buf, &who, 1);
#ifdef DEBUG
	if (ProcessDebug(2))
	    yoke_printf("Manager %d is ready\n", who);
#endif
	yoke_free_buffer(buf);
	yoke_release_transient(sender);
	EndSafeRegion;
    }

    buf = yoke_allocate_buffer(him_conn,
			       yoke_sizeof_conn(him_conn, yoke_self()), 1);
    yoke_put_conn(buf, yoke_self());
    StartSafeRegion;
    yoke_establish(him_conn);
    yoke_send(him_conn, buf, FM_MTYPE_MAIN_RUN);
    yoke_release(him_conn);
    EndSafeRegion;
    
    _fm_manager_run();
    exit(0);
}


/*
 * init_globals()
 *
 * Initialize all the globals to reasonable values.
 */
static void init_globals()
{
    _fm_process_name = (char *) NULL;
    _fm_parent_process_name = (char *) NULL;
    
    _fm_am_manager = TRUE;
    _fm_my_manager = (conn_t) NULL;
    _fm_manager_id = -1;
    _fm_manager_list = (conn_t *) NULL;
    _fm_manager_count = 0;

    _fm_n_close_acks_needed = 0;

    _fm_vp_to_manager_mapping_type = FM_VP_TO_MANAGER_CYCLIC;
    _fm_vp_to_manager_func = (conn_t (*)(int)) NULL;

    _fm_vp_id = 0;
    _fm_vc_offset = 0;
    _fm_vc_base = 0;
    _fm_vc_size = 0;
    _fm_next_channel_id = 0;
    _fm_process_creator = (conn_t) NULL;

    _fm_process_remote_pblock = 0;
    _fm_process_slot = -1;

#ifdef DEBUG
    _fm_debug_level = 0;
    _fm_port_debug_level = 0;
    _fm_migrate_debug_level = 0;
    _fm_process_debug_level = 0;
#endif

    _fm_pause_on_user_startup = FALSE;
    _fm_sleep_on_user_startup = 0;

    _fm_pausepoints = (fm_pausepoint_t) NULL;
    _fm_show_pausepoints = FALSE;

    _fm_init_fortran_io = FALSE;

    _fm_migration_q_head = _fm_migration_q_tail =
	(fm_migration_queue_entry_t) NULL;

    /*
     * Force references to the proper variables.  This will cause
     * ld to fail if the the code being compiled is linked
     * with the incorrect runtime.
     */
#ifdef DEBUG
    _fm_compiled_with_debugging = 1;
    _yoke_compiled_with_debugging = 1;
#else 
    _fm_not_compiled_with_debugging = 1;
    _yoke_not_compiled_with_debugging = 1;
#endif
#ifdef MALLOC_DEBUG
    _fm_compiled_with_malloc_debugging = 1;
    _yoke_compiled_with_malloc_debugging = 1;
#else 
    _fm_not_compiled_with_malloc_debugging = 1;
    _yoke_not_compiled_with_malloc_debugging = 1;
#endif
#ifdef PROFILING
    _fm_compiled_with_profiling = 1;
    _yoke_compiled_with_profiling = 1;
#else
    _fm_not_compiled_with_profiling = 1;
    _yoke_not_compiled_with_profiling = 1;
#endif
}

/*
 * process_args()
 *
 * Process arguments. This gets invoked from yoke_init() with the
 * index of the current argument and the number of arguments.
 */
static int process_args(int current_arg, int arg_count)
{
    char *arg, *mtype;

    arg = _u_get_arg(UTIL_PACKAGE_ARGS, current_arg);
    if ((strcmp(arg, "-maptype") == 0) && ((current_arg + 1) < arg_count))
    {
	mtype = _u_get_arg(UTIL_PACKAGE_ARGS, current_arg + 1);
	if (strcmp(mtype, "cyclic") == 0)
	    _fm_vp_to_manager_mapping_type = FM_VP_TO_MANAGER_CYCLIC;
	else if  (strcmp(mtype, "block") == 0)
	    _fm_vp_to_manager_mapping_type = FM_VP_TO_MANAGER_BLOCK;
	else
	    yoke_fatal("Invalid argument to -maptype\n");
	current_arg += 2;
    }
    else if ((strncmp(arg, "-D", 2) == 0) && ((current_arg + 1) < arg_count) &&
	     ((strcmp(arg+2, "fm") == 0) ||
	      (strcmp(arg+2, "process") == 0) ||
	      (strcmp(arg+2, "migrate") == 0) ||
	      (strcmp(arg+2, "port") == 0)))
	     
    {
	int *var;

#ifndef DEBUG

	    yoke_fatal("The %s option is not supported in the nondebugging library.\nCompile with -g to enable debugging.\n",
		       arg);
#else /* DEBUG */
	
	var = (int *) NULL;

	if (strcmp(arg + 2, "fm") == 0)
	    var = &_fm_debug_level;
	else if (strcmp(arg + 2, "port") == 0)
	    var = &_fm_port_debug_level;
	else if (strcmp(arg + 2, "migrate") == 0)
	    var = &_fm_migrate_debug_level;
	else if (strcmp(arg + 2, "process") == 0)
	    var = &_fm_process_debug_level;

	if (var != (int *) NULL)
	{
	    *var = atoi(_u_get_arg(UTIL_PACKAGE_ARGS, current_arg + 1));
	    current_arg += 2;
	}
#endif /* DEBUG */
    }
    else if (strcmp(arg, "-pause_on_user_startup") == 0)
    {
	_fm_pause_on_user_startup = TRUE;
	current_arg++;
    }
    else if ((strcmp(arg, "-sleep_on_user_startup") == 0) &&
	     ((current_arg + 1) < arg_count))	     
    {
	_fm_sleep_on_user_startup = atoi(_u_get_arg(UTIL_PACKAGE_ARGS,
						    current_arg + 1));;
	current_arg += 2;
    }
    else if ((strcmp(arg, "-pause") == 0) &&
	     ((current_arg + 1) < arg_count))	     
    {
	_fm_add_pausepoint(atoi(_u_get_arg(UTIL_PACKAGE_ARGS, current_arg + 1)));
	current_arg += 2;
    }
    else if ((strcmp(arg, "-sleep") == 0) &&
	     ((current_arg + 2) < arg_count))	     
    {
	_fm_add_sleeppoint(atoi(_u_get_arg(UTIL_PACKAGE_ARGS, current_arg + 1)),
			   atoi(_u_get_arg(UTIL_PACKAGE_ARGS, current_arg + 2)));
	current_arg += 3;
    }
    else if (strcmp(arg, "-show_pausepoints") == 0)
    {
	_fm_show_pausepoints = TRUE;
	current_arg += 1;
    }
#ifdef NOTDEF
    else if (strcmp(arg, "-init_fortran_io") == 0)
    {
	_fm_init_fortran_io = TRUE;
	current_arg++;
    }
#endif
    return current_arg;
}


static void show_usage_message()
{
    printf("    -maptype <type>           : Set the physical-to-virtual mapping type\n");
    printf("                                to <type>. Valid types are cyclic and block.\n");
    printf("                                The default is cyclic.\n");
    printf("    -Dfm <level>              : Set the general debug level to <level>\n");
    printf("    -Dprocess <level>         : Set the process debug level to <level>\n");
    printf("    -Dmigrate <level>         : Set the migration debug level to <level>\n");
    printf("    -Dport <level>            : Set the port debug level to <level>\n");
    printf("    -pause_on_user_startup    : Execute the pause() system call just before\n");
    printf("                                entering user code.\n");
    printf("    -sleep_on_user_startup <n>: Sleep for <n> seconds just before\n");
    printf("                                entering user code.\n");
    printf("    -pause <id>               : Pause at pausepoint <id>\n");
    printf("    -sleep <id> <n>           : Sleep for <n> seconds at pausepoint <id>\n");
    printf("    -show_pausepoints         : Print a message when each pausepoint is \n");
    printf("                                executed\n");
}

/*
 * set_vp_to_manager_mapping_type()
 *
 * Set the value of the _fm_vp_to_manager_func variable to the
 * appropriate function based on the value of _fm_vp_to_manager_mapping_type.
 *
 * We do this here instead of directly setting it because each remote
 * node must set it based on the message received from the master.
 */
static void set_vp_to_manager_mapping_type()
{
    switch (_fm_vp_to_manager_mapping_type)
    {
    case FM_VP_TO_MANAGER_BLOCK:
	_fm_vp_to_manager_func = _fm_vp_to_manager_block;
	break;

    case FM_VP_TO_MANAGER_CYCLIC:
	_fm_vp_to_manager_func = _fm_vp_to_manager_cyclic;
	break;

    default:
	yoke_fatal("fm_init(): Invalid VP to manager mapping type %d\n",
		   _fm_vp_to_manager_mapping_type);
    }
}

/*
 * Fortran M remote node intialization.
 *
 * We receive the VC table from the master, initialize the manager
 * data structures, and just drop into the manager main loop.
 */
static void node_main()
{
    yoke_buffer_t buf;
    conn_t dest;
    
    recv_initialization_from_master();

    set_vp_to_manager_mapping_type();

    _fm_manager_init();

    /*
     * We're all set. Tell the master we are ready to go.
     */

    dest = _fm_manager_list[0];
    buf = yoke_allocate_buffer(dest,
			       yoke_sizeof_int(dest, 1),
			       1);
    yoke_put_int(buf, &_fm_manager_id, 1);
    StartSafeRegion;
    yoke_establish(dest);
    yoke_send(dest, buf, FM_MTYPE_MANAGER_READY);
    yoke_release(dest);
    EndSafeRegion;

    _fm_manager_run();
}


/*
 * send_initialization_to_nodes()
 *
 * Send off any information the nodes will need. This currently includes
 *
 *    (int)	Debug flag value
 *    (int)	VP-to-manager mapping type
 *    (int)	Virtual machine size
 *    (int)	Manager list size
 *    (conn_t)	Manager list
 *    (int)     Fortran I/O initialization flag
 */
static void send_initialization_to_nodes()
{
    int i, j, size, count, ppsize, ppcount;
    yoke_buffer_t buf;
    conn_t dest;

    for (i = 1; i < _fm_manager_count; i++)
    {
	dest = _fm_manager_list[i];

	size = yoke_sizeof_int(dest, 6);
	count = 6;

	_fm_sizeof_pausepoints(dest, &ppsize, &ppcount);
	size += ppsize;
	count += ppcount;
	
#ifdef DEBUG
	size += yoke_sizeof_int(dest, 4);
	count += 4;
#endif
	for (j = 0; j < _fm_manager_count; j++)
	{
	    count++;
	    size += yoke_sizeof_conn(dest, _fm_manager_list[j]);
	}
	
	buf = yoke_allocate_buffer(dest, size, count);

	_fm_put_pausepoints(buf);
	yoke_put_int(buf, &_fm_show_pausepoints, 1);

#ifdef DEBUG
	yoke_put_int(buf, &_fm_debug_level, 1);
	yoke_put_int(buf, &_fm_port_debug_level, 1);
	yoke_put_int(buf, &_fm_migrate_debug_level, 1);
	yoke_put_int(buf, &_fm_process_debug_level, 1);
#endif
	yoke_put_int(buf, &i, 1);
	yoke_put_int(buf, &_fm_vp_to_manager_mapping_type, 1);
	yoke_put_int(buf, &_fm_machine_size, 1);
	yoke_put_int(buf, &_fm_manager_count, 1);
	yoke_put_int(buf, &_fm_init_fortran_io, 1);
	for (j = 0; j < _fm_manager_count; j++)
	   yoke_put_conn(buf, _fm_manager_list[j]);
	StartSafeRegion;
	yoke_establish(dest);
	yoke_send(dest, buf, FM_MTYPE_NODE_INITIALIZATION);
	yoke_release(dest);
	EndSafeRegion;
    }
}

/*
 * recv_initialization_from_master()
 *
 * Receive the initialization information as sent by
 * send_initialization_to_nodes().
 */
static void recv_initialization_from_master()
{
    yoke_buffer_t buf;
    conn_t sender;
    int i, type;

    StartSafeRegion;
    buf = yoke_recv_transient(&sender, &type, FALSE);
    if (type != FM_MTYPE_NODE_INITIALIZATION)
	yoke_fatal("recv_initialization_from_master(): expected type FM_MTYPE_NODE_INITIALIZATION\nbut got %d\n", type);

    yoke_release_transient(sender);
    EndSafeRegion;

    _fm_get_pausepoints(buf);
    yoke_get_int(buf, &_fm_show_pausepoints, 1);

#ifdef DEBUG
    yoke_get_int(buf, &_fm_debug_level, 1);
    yoke_get_int(buf, &_fm_port_debug_level, 1);
    yoke_get_int(buf, &_fm_migrate_debug_level, 1);
    yoke_get_int(buf, &_fm_process_debug_level, 1);
#endif
    yoke_get_int(buf, &_fm_manager_id, 1);
    yoke_get_int(buf, &_fm_vp_to_manager_mapping_type, 1);
    yoke_get_int(buf, &_fm_machine_size, 1);
    yoke_get_int(buf, &_fm_manager_count, 1);
    yoke_get_int(buf, &_fm_init_fortran_io, 1);

    _fm_manager_list = (conn_t *) malloc(sizeof(conn_t) * _fm_manager_count);
    if (_fm_manager_list == (conn_t *) NULL)
	yoke_fatal("recv_initialization_from_master(): malloc failed\n");
    
    for (i = 0; i < _fm_manager_count; i++)
	yoke_get_conn(buf, &(_fm_manager_list[i]));

    yoke_free_buffer(buf);
}
