/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */

#include "fm.h"

/*
 * Port indices.
 *
 * We maintain here a hashtable mapping from channel/id pairs to
 * the port variables that refer to them.
 *
 */

typedef struct _pindex_entry_t
{
    conn_t conn;
    int id;
    
    fm_inport_t inport;
    fm_outport_t *outports;

    int n_outports, n_outports_alloc;

    fm_inport_migrec_t inport_migrec;

    fm_outport_migrec_t *outport_migrecs;
    int n_outport_migrecs, n_outport_migrecs_alloc;

    struct _pindex_entry_t *next;
} *pindex_entry_t;

#define PORT_INDEX_SIZE 17

static struct _pindex_entry_t pindex_table[PORT_INDEX_SIZE];

static unsigned int gen_hash(int id, conn_t conn);
static void add_inport_to_entry(pindex_entry_t entry, fm_inport_t inport);
static void add_outport_to_entry(pindex_entry_t entry, fm_outport_t outport);
static void remove_inport_from_entry(pindex_entry_t entry, fm_inport_t inport);
static void remove_outport_from_entry(pindex_entry_t entry,
				      fm_outport_t outport);

static void remove_outport_migrec_from_entry(pindex_entry_t entry,
					     fm_outport_migrec_t outport_migrec);
static void remove_inport_migrec_from_entry(pindex_entry_t entry,
					    fm_inport_migrec_t inport_migrec);
static void add_outport_migrec_to_entry(pindex_entry_t entry,
					fm_outport_migrec_t outport_migrec);
static void add_inport_migrec_to_entry(pindex_entry_t entry,
				       fm_inport_migrec_t inport_migrec);


static pindex_entry_t lookup_entry(int id, conn_t conn);
static pindex_entry_t lookup_entry_create(int id, conn_t conn);

/*
 * _fm_pindex_init()
 *
 * Clear out the hashtable. 
 */
void _fm_pindex_init()
{
    ZeroOutMemory(pindex_table, sizeof(pindex_table));
}


/*
 * _fm_pindex_add_outport()
 *
 */
void _fm_pindex_add_outport(fm_outport_t outport)
{
    pindex_entry_t ent;
    
    FM_OUTPORT_MAGIC_CHECK(_fm_pindex_add_outport, outport);

#ifdef DEBUG
    if (MigrateDebug(3))
	yoke_printf("_fm_pindex_add_outport(): outport=%s\n",
		    _fm_outport_as_string(outport));
#endif
    
    if (outport->channel_id < 0)
	return;

    ent = lookup_entry_create(outport->channel_id, outport->channel_conn);
    Assert(ent != (pindex_entry_t) NULL);
    add_outport_to_entry(ent, outport);
}

/*
 * _fm_pindex_add_inport()
 *
 */
void _fm_pindex_add_inport(fm_inport_t inport)
{
    pindex_entry_t ent;
    
    FM_INPORT_MAGIC_CHECK(_fm_pindex_add_inport, inport);

#ifdef DEBUG
    if (MigrateDebug(3))
	yoke_printf("_fm_pindex_add_inport(): inport=%s\n",
		    _fm_inport_as_string(inport));
#endif
    
    if (inport->channel_id < 0)
	return;

    ent = lookup_entry_create(inport->channel_id, inport->channel_conn);
    Assert(ent != (pindex_entry_t) NULL);
    add_inport_to_entry(ent, inport);
}


/*
 * _fm_pindex_remove_outport()
 *
 */
void _fm_pindex_remove_outport(fm_outport_t outport)
{
    pindex_entry_t ent;
    
    FM_OUTPORT_MAGIC_CHECK(_fm_pindex_remove_outport, outport);

#ifdef DEBUG
    if (MigrateDebug(3))
	yoke_printf("_fm_pindex_remove_outport(): outport=%s\n",
		    _fm_outport_as_string(outport));
#endif
    
    if (outport->channel_id < 0)
	return;

    ent = lookup_entry(outport->channel_id, outport->channel_conn);
    if (ent == (pindex_entry_t) NULL)
	yoke_fatal("_fm_pindex_remove_outport(): Outport %s has no entry\n",
		   PortName(outport));

    remove_outport_from_entry(ent, outport);
}


/*
 * _fm_pindex_remove_outport_with_channel()
 *
 */
void _fm_pindex_remove_outport_with_channel(fm_outport_t outport,
					    conn_t channel_conn,
					    int channel_id)
{
    pindex_entry_t ent;
    
    FM_OUTPORT_MAGIC_CHECK(_fm_pindex_remove_outport_with_channel, outport);

#ifdef DEBUG
    if (MigrateDebug(3))
	yoke_printf("_fm_pindex_remove_outport_with_channel(): outport=%s conn=%s id=%d\n",
		    _fm_outport_as_string(outport),
		    yoke_conn_as_string(channel_conn),
		    channel_id);
#endif
    
    if (channel_id < 0)
	return;

    ent = lookup_entry(channel_id, channel_conn);
    if (ent == (pindex_entry_t) NULL)
	yoke_fatal("_fm_pindex_remove_outport(): Outport %s has no entry\n",
		   PortName(outport));

    remove_outport_from_entry(ent, outport);
}


/*
 * _fm_pindex_remove_inport()
 *
 */
void _fm_pindex_remove_inport(fm_inport_t inport)
{
    pindex_entry_t ent;
    
    FM_INPORT_MAGIC_CHECK(_fm_pindex_remove_inport, inport);

#ifdef DEBUG
    if (MigrateDebug(3))
	yoke_printf("_fm_pindex_remove_inport(): inport=%s\n",
		    _fm_inport_as_string(inport));
#endif

    if (inport->channel_id < 0)
	return;

    ent = lookup_entry(inport->channel_id, inport->channel_conn);
    if (ent == (pindex_entry_t) NULL)
	yoke_fatal("_fm_pindex_remove_inport(): Inport %s has no entry\n",
		   PortName(inport));

    remove_inport_from_entry(ent, inport);
}


/*
 * _fm_pindex_remove_inport_with_channel()
 *
 */
void _fm_pindex_remove_inport_with_channel(fm_inport_t inport,
					   conn_t channel_conn,
					   int channel_id)
{
    pindex_entry_t ent;
    
    FM_INPORT_MAGIC_CHECK(_fm_pindex_remove_inport_with_channel, inport);

#ifdef DEBUG
    if (MigrateDebug(3))
	yoke_printf("_fm_pindex_remove_inport_with_channel(): inport=%s conn=%s id=%d\n",
		    _fm_inport_as_string(inport),
		    yoke_conn_as_string(channel_conn),
		    channel_id);
#endif

    if (channel_id < 0)
	return;

    ent = lookup_entry(channel_id, channel_conn);
    if (ent == (pindex_entry_t) NULL)
	yoke_fatal("_fm_pindex_remove_inport_with_channel(): Inport %s has no entry\n",
		   PortName(inport));

    remove_inport_from_entry(ent, inport);
}

/*
 * Functions for adding and removing migration records.
 */

void _fm_pindex_add_inport_migrec(conn_t conn, int id,
				  fm_inport_migrec_t migrec)
{
    pindex_entry_t ent;
    
#ifdef DEBUG
    if (MigrateDebug(3))
	yoke_printf("_fm_pindex_add_inport_migrec(): migrec=%x conn=%s id=%d\n",
		    migrec, yoke_conn_as_string(conn), id);
#endif

    ent = lookup_entry_create(id, conn);
    Assert(ent != (pindex_entry_t) NULL);
    add_inport_migrec_to_entry(ent, migrec);
}


void _fm_pindex_add_outport_migrec(conn_t conn, int id,
				  fm_outport_migrec_t migrec)
{
    pindex_entry_t ent;
    
#ifdef DEBUG
    if (MigrateDebug(3))
	yoke_printf("_fm_pindex_add_outport_migrec(): migrec=%x conn=%s id=%d\n",
		    migrec, yoke_conn_as_string(conn), id);
#endif

    ent = lookup_entry_create(id, conn);
    Assert(ent != (pindex_entry_t) NULL);
    add_outport_migrec_to_entry(ent, migrec);
}


/*
 * _fm_pindex_lookup_ports()
 *
 * The FM entry point for looking up ports in the index.
 *
 */
void _fm_pindex_lookup_ports(conn_t conn, int id,
			     fm_inport_t *inport,
			     fm_outport_t **outports,
			     int *n_outports)
{
    pindex_entry_t ent;

    ent = lookup_entry(id, conn);
    if (ent == (pindex_entry_t) NULL)
    {
	*inport = (fm_inport_t) NULL;
	*outports = (fm_outport_t *) NULL;
	*n_outports = 0;
    }
    else
    {
	*inport = ent->inport;
	*outports = ent->outports;
	*n_outports = ent->n_outports;
    }
}

/*
 * _fm_pindex_lookup_migrecs()
 *
 * The FM entry point for looking up migrecs in the index.
 *
 */
void _fm_pindex_lookup_migrecs(conn_t conn, int id,
			     fm_inport_migrec_t *inport_migrec,
			     fm_outport_migrec_t **outport_migrecs,
			     int *n_outport_migrecs)
{
    pindex_entry_t ent;

    ent = lookup_entry(id, conn);
    if (ent == (pindex_entry_t) NULL)
    {
	*inport_migrec = (fm_inport_migrec_t) NULL;
	*outport_migrecs = (fm_outport_migrec_t *) NULL;
	*n_outport_migrecs = 0;
    }
    else
    {
	*inport_migrec = ent->inport_migrec;
	*outport_migrecs = ent->outport_migrecs;
	*n_outport_migrecs = ent->n_outport_migrecs;
    }
}

void _fm_pindex_remove_inport_migrec(conn_t conn, int id,
				     fm_inport_migrec_t migrec)
{
    pindex_entry_t ent;
    
#ifdef DEBUG
    if (MigrateDebug(3))
	yoke_printf("_fm_pindex_remove_inport_migrec(): migrec=%x conn=%s id=%d\n",
		    migrec, yoke_conn_as_string(conn), id);
#endif

    ent = lookup_entry(id, conn);

    if (ent == (pindex_entry_t) NULL)
	yoke_fatal("_fm_pindex_remove_inport_migrec(): Inport migrec %x has no entry\n",
		   migrec);


    remove_inport_migrec_from_entry(ent, migrec);
}


void _fm_pindex_remove_outport_migrec(conn_t conn, int id,
				     fm_outport_migrec_t migrec)
{
    pindex_entry_t ent;
    
#ifdef DEBUG
    if (MigrateDebug(3))
	yoke_printf("_fm_pindex_remove_outport_migrec(): migrec=%x conn=%s id=%d\n",
		    migrec, yoke_conn_as_string(conn), id);
#endif

    ent = lookup_entry(id, conn);

    if (ent == (pindex_entry_t) NULL)
	yoke_fatal("_fm_pindex_remove_outport_migrec(): Outport migrec %x has no entry\n",
		   migrec);


    remove_outport_migrec_from_entry(ent, migrec);
}


static pindex_entry_t lookup_entry(int id, conn_t conn)
{
    unsigned int hval;
    pindex_entry_t ent;

    hval = gen_hash(id, conn);

    for (ent = pindex_table[hval].next; ent != (pindex_entry_t) NULL;
	 ent = ent->next)
    {
	if (ent->id == id && yoke_conn_equal(ent->conn, conn))
	    return ent;
    }
    return (pindex_entry_t) NULL;
}


static pindex_entry_t lookup_entry_create(int id, conn_t conn)
{
    unsigned int hval;
    pindex_entry_t ent;

    hval = gen_hash(id, conn);

    for (ent = pindex_table[hval].next; ent != (pindex_entry_t) NULL;
	 ent = ent->next)
    {
	if (ent->id == id && yoke_conn_equal(ent->conn, conn))
	    return ent;
    }
    ent = (pindex_entry_t) malloc(sizeof(struct _pindex_entry_t));
    MallocCheck(lookup_entry_create(), ent, pindex_entry_t);

    ent->conn = conn;
    ent->id = id;

    ent->inport = (fm_inport_t) NULL;
    ent->outports = (fm_outport_t *) NULL;

    ent->n_outports = ent->n_outports_alloc = 0;

    ent->inport_migrec = (fm_inport_migrec_t) NULL;

    ent->outport_migrecs = (fm_outport_migrec_t *) NULL;
    ent->n_outport_migrecs = ent->n_outport_migrecs_alloc = 0;

    ent->next = pindex_table[hval].next;
    
    pindex_table[hval].next = ent;
    
    return (pindex_entry_t) ent;
}


static unsigned int gen_hash(int id, conn_t conn)
{
    int pid;
    char *host, *s;
    unsigned int hval;

    hval = 0;
    yoke_get_conn_id(conn, &host, &pid);

    hval = pid;
    
    for (s = host; *s != '\0'; s++)
    {
	hval = (hval << 1) + *s;
    }

    return hval % PORT_INDEX_SIZE;
}


static void add_inport_to_entry(pindex_entry_t entry, fm_inport_t inport)
{
    if (entry->inport != (fm_inport_t) NULL)
    {
	FM_INPORT_MAGIC_CHECK(add_inport_to_entry, entry->inport);
	
	yoke_fatal("add_inport_to_entry(): Entry %x %s/%d already has an inport %s (adding %s)\n",
		   entry,
		   yoke_conn_as_string(entry->conn),
		   entry->id,
		   _fm_inport_as_string(entry->inport),
		   _fm_inport_as_string(inport));
    }

    entry->inport = inport;
}


static void add_inport_migrec_to_entry(pindex_entry_t entry,
				       fm_inport_migrec_t inport_migrec)
{
    if (entry->inport_migrec != (fm_inport_migrec_t) NULL)
    {
	yoke_fatal("add_inport_migrec_to_entry(): Entry %x %s/%d already has an inport_migrec %x (adding %x)\n",
		   entry,
		   yoke_conn_as_string(entry->conn),
		   entry->id,
		   entry->inport_migrec,
		   inport_migrec);
    }

    entry->inport_migrec = inport_migrec;
}


static void add_outport_to_entry(pindex_entry_t entry, fm_outport_t outport)
{
    if (entry->n_outports == entry->n_outports_alloc ||
	entry->n_outports_alloc == 0)
    {
	int new_size, i;
	fm_outport_t *new_list;

	if (entry->n_outports_alloc == 0)
	    new_size = 5;
	else
	    new_size = entry->n_outports_alloc * 2;

#ifdef DEBUG
	if (MigrateDebug(3))
	    yoke_printf("Allocating new outport list of size %d\n", new_size);
#endif
	
	new_list = (fm_outport_t *) malloc(sizeof(fm_outport_t) * new_size);
	MallocCheck(add_outport_to_entry(), new_list, fm_outport_t *);

#ifdef DEBUG
	if (entry->n_outports > 0)
	    Assert(entry->outports != (fm_outport_t *) NULL);
#endif

	for (i = 0; i < entry->n_outports; i++)
	{
	    FM_OUTPORT_MAGIC_CHECK(add_outport_to_entry, entry->outports[i]);
	    new_list[i] = entry->outports[i];
	}

	if (entry->outports)
	    free(entry->outports);

	entry->outports = new_list;
	entry->n_outports_alloc = new_size;
    }
    entry->outports[entry->n_outports++] = outport;
}

static void add_outport_migrec_to_entry(pindex_entry_t entry,
					fm_outport_migrec_t outport_migrec)
{
    if (entry->n_outport_migrecs == entry->n_outport_migrecs_alloc ||
	entry->n_outport_migrecs_alloc == 0)
    {
	int new_size, i;
	fm_outport_migrec_t *new_list;

	if (entry->n_outport_migrecs_alloc == 0)
	    new_size = 5;
	else
	    new_size = entry->n_outport_migrecs_alloc * 2;

#ifdef DEBUG
	if (MigrateDebug(3))
	    yoke_printf("Allocating new outport_migrec list of size %d\n", new_size);
#endif
	
	new_list = (fm_outport_migrec_t *) malloc(sizeof(fm_outport_migrec_t) * new_size);
	MallocCheck(add_outport_migrec_to_entry(), new_list, fm_outport_migrec_t *);

#ifdef DEBUG
	if (entry->n_outport_migrecs > 0)
	    Assert(entry->outport_migrecs != (fm_outport_migrec_t *) NULL);
#endif

	for (i = 0; i < entry->n_outport_migrecs; i++)
	{
	    new_list[i] = entry->outport_migrecs[i];
	}

	if (entry->outport_migrecs)
	    free(entry->outport_migrecs);

	entry->outport_migrecs = new_list;
	entry->n_outport_migrecs_alloc = new_size;
    }
    entry->outport_migrecs[entry->n_outport_migrecs++] = outport_migrec;
}


static void remove_inport_from_entry(pindex_entry_t entry, fm_inport_t inport)
{
    if (entry->inport == (fm_inport_t) NULL)
    {
	FM_DUMP_PORT_INDEX("before fatal", 11);
	yoke_fatal("remove_inport_from_entry(): Entry has no inport while removing %s\n",
		   _fm_inport_as_string(inport));
    }

    FM_INPORT_MAGIC_CHECK(remove_inport_for_entry, entry->inport);

    if (inport->channel_id != entry->inport->channel_id ||
	!yoke_conn_equal(inport->channel_conn, entry->inport->channel_conn))
	yoke_fatal("remove_inport_from_entry(): inport %s not equal to entry inport %s\n",
		   _fm_inport_as_string(inport),
		   _fm_inport_as_string(entry->inport));

    entry->inport = (fm_inport_t) NULL;
}


static void remove_inport_migrec_from_entry(pindex_entry_t entry,
					    fm_inport_migrec_t inport_migrec)
{
    if (entry->inport_migrec == (fm_inport_migrec_t) NULL)
    {
	FM_DUMP_PORT_INDEX("before fatal", 11);
	yoke_fatal("remove_inport_migrec_from_entry(): Entry has no inport_migrec while removing %s\n",
		   inport_migrec);
    }

    if (inport_migrec != entry->inport_migrec)
	yoke_fatal("remove_inport_from_entry(): inport %x not equal to entry inport %x\n",
		   inport_migrec, entry->inport_migrec);

    entry->inport_migrec = (fm_inport_migrec_t) NULL;
}


static void remove_outport_from_entry(pindex_entry_t entry,
				      fm_outport_t outport)
{
    int i;
    int remove_idx;
    fm_outport_t e_outport;
    
    if (entry->outports == (fm_outport_t *) NULL)
	yoke_fatal("remove_outport_from_entry(): Entry has no outports while removing %s\n",
		   _fm_outport_as_string(outport));

    remove_idx = -1;
    for (i = 0; i < entry->n_outports; i++)
    {
	e_outport = entry->outports[i];
	FM_OUTPORT_MAGIC_CHECK(remove_outport_from_entry, e_outport);
	
	if (outport->channel_id == e_outport->channel_id &&
	    outport->inport_index == e_outport->inport_index && 
	    yoke_conn_equal(outport->channel_conn, e_outport->channel_conn))
	{
	    remove_idx = i;
	    break;
	}
    }
    if (remove_idx == -1)
    {
	FM_DUMP_PORT_INDEX("before fatal", 11);
	yoke_fatal("remove_outport_from_entry(): Did not find outport %s in entry\n",
		   _fm_outport_as_string(outport));
    }

    for (i = remove_idx; i < entry->n_outports - 1; i++)
    {
	entry->outports[i] = entry->outports[i + 1];
    }

    entry->n_outports--;
}

static void remove_outport_migrec_from_entry(pindex_entry_t entry,
				      fm_outport_migrec_t outport_migrec)
{
    int i;
    int remove_idx;
    
    if (entry->outport_migrecs == (fm_outport_migrec_t *) NULL)
	yoke_fatal("remove_outport_migrec_from_entry(): Entry has no outport_migrecs while removing %x\n",
		   outport_migrec);


    for (i = 0; i < entry->n_outport_migrecs; i++)
    {
	if (outport_migrec == entry->outport_migrecs[i])
	{
	    remove_idx = i;
	    break;
	}
    }

    if (remove_idx == -1)
    {
	FM_DUMP_PORT_INDEX("before fatal", 12);
	yoke_fatal("remove_outport_migrec_from_entry(): Did not find outport_migrec %x in entry\n",
		   outport_migrec);

    }

    for (i = remove_idx; i < entry->n_outport_migrecs - 1; i++)
    {
	entry->outport_migrecs[i] = entry->outport_migrecs[i + 1];
    }

    entry->n_outport_migrecs--;
}


/*
 * FM_DUMP_PORT_INDEX()
 *
 * Debugging routine, callable from Fortran.
 *
 * Dumps this process' port index.
 */
void FM_DUMP_PORT_INDEX(char *str, int len)
{
    int i, j;
    pindex_entry_t ent;
    char string[1024];

#ifdef DEBUG
    if (MigrateDebug(2))
    {
        if (len > sizeof(string)-1)
        {
            strncpy(string, str, sizeof(string)-1);
            string[sizeof(string)-1] = '\0';
        }
        else
        {
            strncpy(string, str, len);
            string[len] = '\0';
        }
        
        yoke_printf("Port index dump: %s\n", string);
        for (i = 0; i < PORT_INDEX_SIZE; i++)
        {
            for (ent = pindex_table[i].next; ent != (pindex_entry_t) NULL;
                 ent = ent->next)
            {
                yoke_printf("bucket=%d ent=%x conn=%s id=%d\n",
                            i, ent,
                            yoke_conn_as_string(ent->conn),
			    ent->id);
		if (ent->inport == (fm_inport_t) NULL)
		{
		    yoke_printf("inport=(none)\n");
		}
		else
		{
		    FM_INPORT_MAGIC_CHECK(FM_DUMP_PORT_INDEX, ent->inport);
		    yoke_printf("inport=%s\n",
				ent->inport == (fm_inport_t) NULL ? "(none)" :
				_fm_inport_as_string(ent->inport));
		}
		
		yoke_printf("n_outports=%d n_outports_alloc=%d\n",
			    ent->n_outports,
			    ent->n_outports_alloc);
		
		for (j = 0; j < ent->n_outports; j++)
		{
		    FM_OUTPORT_MAGIC_CHECK(FM_DUMP_PORT_INDEX,
					   ent->outports[j]);
		    yoke_printf("  outports[%d]=%s\n",
				j, _fm_outport_as_string(ent->outports[j]));
		}

		yoke_printf("    inport_migrec=%x\n", ent->inport_migrec);
		yoke_printf("n_outport_migrecs=%d n_outport_migrecs_alloc=%d next=%x\n",
			    ent->n_outport_migrecs,
			    ent->n_outport_migrecs_alloc,
			    ent->next);
		
		for (j = 0; j < ent->n_outport_migrecs; j++)
		{
		    yoke_printf("  outport_migrecs[%d]=%x\n",
				j, ent->outport_migrecs[j]);
		}
	    }
        }
    }
#endif
}
